/*
Inspired of Denis V. Chumakov Simple SWF Parser
http://flashpanoramas.com/blog/

Use this code without any restrictions.
*/

package
{
	import flash.display.Graphics;
	import flash.display.Sprite;
	import flash.display.StageAlign;
	import flash.display.StageScaleMode;
	import flash.events.Event;
	import flash.events.HTTPStatusEvent;
	import flash.events.IOErrorEvent;
	import flash.events.MouseEvent;
	import flash.events.ProgressEvent;
	import flash.events.SecurityErrorEvent;
	import flash.geom.Rectangle;
	import flash.net.URLRequest;
	import flash.net.URLLoader;
	import flash.net.URLLoaderDataFormat;
	import flash.system.System;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import flash.text.TextFormat;
	import flash.utils.ByteArray;
	import flash.utils.Endian;
	import flash.utils.getTimer;

	public class SWFReader extends Sprite
	{
		// handlers for SWF tags
		// for example: 
		// handlers[6] = parseJPEG; 
		// to set your handler for DefineBitsJPEG tag.
		// you can find SWF file specifications in Google.
		private var handlers:Array = [];
		private static const SWF_TAGS_NAME:Array = [
			"End",//0
			"ShowFrame",
			"DefineShape",
			"FreeCharacter ?",
			"PlaceObject",
			"RemoveObject",
			"DefineBitsJPEG",
			"DefineButton",
			"JPEGTables",
			"SetBackgroundColor",
			"DefineFont",//10
			"DefineText",
			"DoAction",
			"DefineFontInfo",
			"DefineSound",
			"StartSound",
			"StopSound ?",
			"DefineButtonSound",
			"SoundStreamHead",
			"SoundStreamBlock",
			"DefineBitsLossless",//20
			"DefineBitsJPEG2",
			"DefineShape2",
			"DefineButtonCxform",
			"Protect",
			"PathsArePostscript ?",
			"PlaceObject2",//26
			"?",
			"RemoveObject2",//28
			"SyncFrame ?",//29
			"?",
			"FreeAll ?",//31
			"DefineShape3",
			"DefineText2",
			"DefineButton2",
			"DefineBitsJPEG3",
			"DefineBitsLossless2",//36
			"DefineEditText",
			"DefineVideo ?",
			"DefineSprite",//39
			"NameCharacter ?",//40
			"ProductInfo",
			"DefineTextFormat ?",
			"FrameLabel",//43
			"?",
			"SoundStreamHead2",//45
			"DefineMorphShape",
			"GenerateFrame ?",
			"DefineFont2",
			"GeneratorCommand",
			"DefineCommandObject ?",//50
			"CharacterSet ?",
			"ExternalFont ?",//52
			"?",
			"?",
			"?",
			"Export",//56
			"Import",
			"ProtectDebug",
			"DoInitAction",
			"DefineVideoStream",
			"VideoFrame",
			"DefineFontInfo2",
			"DebugID",
			"ProtectDebug2",
			"ScriptLimits",
			"SetTabIndex",//66
			"?",
			"?",
			"FileAttributes",//69
			"PlaceObject3",//70
			"Import2",
			"DoABC",
			"DefineFontAlignZones",
			"CSMTextSettings",
			"DefineFont3",
			"SymbolClass",
			"Metadata",
			"DefineScalingGrid",//78
			"?",
			"?",
			"?",
			"DoABCDefine",//82
			"DefineShape4",
			"DefineMorphShape2",//84
			"?",
			"DefineSceneAndFrameLabelData",
			"DefineBinaryData",
			"DefineFontName"//88
		];
		private static const LANGUAGE_CODE:Array = [
			"none",
			"Latin",
			"Japanese",
			"Korean",
			"Simplified Chinese",
			"Traditional Chinese"
		];
		
		private var loader:URLLoader = new URLLoader();
		private var request:URLRequest;
		//private static const FILE:String = "SegoeUI.swf";
		private static const FILE:String = "DINCondensedC.swf";
		
		private var output:String = "";
		private var console:TextField = new TextField();
		
		private var numGlyphs:uint = 0;
	
		public function SWFReader():void
		{
			this.stage.align = StageAlign.TOP_LEFT;
			this.stage.scaleMode = StageScaleMode.NO_SCALE;
			
			this.console.autoSize = TextFieldAutoSize.LEFT;
			this.console.width = this.stage.stageWidth;
			this.console.embedFonts = true;
			this.console.defaultTextFormat = new TextFormat("CodingFontTobi", 16);
			this.addChild(this.console);
			
			this.loader.dataFormat = URLLoaderDataFormat.BINARY;
			this.loader.addEventListener(Event.COMPLETE, this.swfLoadComplete, false, 0, true);
			this.loader.addEventListener(HTTPStatusEvent.HTTP_STATUS, this.checkLoadHTTPStatus, false, 0, true);
			this.loader.addEventListener(IOErrorEvent.IO_ERROR, this.ioErrorHandler, false, 0, true);
			this.loader.addEventListener(ProgressEvent.PROGRESS, this.loadProgressHandler, false, 0, true);
			this.loader.addEventListener(SecurityErrorEvent.SECURITY_ERROR, this.securityErrorHandler, false, 0, true);
			this.request = new URLRequest(this.loaderInfo.parameters.file || FILE);
			this.loader.load(this.request);
		}
		
		private function loadProgressHandler(event:ProgressEvent):void
		{
			this.console.text = uint(event.bytesLoaded / event.bytesTotal) + "%";
		}
		
		private function checkLoadHTTPStatus(event:HTTPStatusEvent):void
		{
			if ((event.status >= 400 || event.status < 300) && event.status != 0)
			{
				this.console.text = "Loading fail\nBad HTTP status: " + event.status;
				/*
				try
				{
					this.loader.close();
				}
				catch (error:Error)
				{
					
				};*/
			}
		}
		
		private function securityErrorHandler(event:SecurityErrorEvent):void
		{
			this.console.text = "Loading fail\nSecurity error: " + event.text;
		}
		
		private function ioErrorHandler(event:IOErrorEvent):void
		{
			this.console.text = "Loading fail\nI/O error: " + event.text;
		}
		
		private function log(name:String, value:Object = null, level:uint = 0):void
		{
			var output:String = "";
			var shift:String = "";
			for (var i:uint = 0; i < level; i ++)
				shift += "\t";
			
			output += shift + name + ":";
			shift += "\t";
			
			if (value == null)
				void;
			else if (value is Array)
				output += "\n" + shift + "[" + (value as Array).join(",") + "]";
			else if (value is ByteArray)
			{
				var byteArray:ByteArray = value as ByteArray;
				var stringRep:String = "";
				var byte:int;
				var i:uint = 0;
				byteArray.position = 0;
				//for (var i:uint = 0; i < byteArray.length; i++) {
				output += "\n" + shift + "Address   +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +a +b +c +d +e +f  Dump";
				while (byteArray.bytesAvailable)
				{
					if (i % 16 == 0)// dump offset
						output += "\n" + shift + ("00000000" + i.toString(16)).substr( -8, 8) + "  ";
					
					/*if (i % 8 == 0)
						output += " ";*/
					
					byte = byteArray.readByte();//value[i];
					output += ("0" + byte.toString(16)).substr(-2, 2) + " ";
					stringRep += (byte < 32 || byte > 126) ? "." : String.fromCharCode(byte);
					//if (((i + 1) % 16 == 0) || (i == byteArray.length - 1))
					if (((i + 1) % 16 == 0))
					{
						output += " " + stringRep;
						stringRep = "";
					}
					if (((i + 1) % 16 != 0) && (i == byteArray.length - 1))
					{
						output += "                                             ".substr(0, 3 * (16 - (i + 1) % 16)) + " " + stringRep;
					}
					
					i ++;
				}
			}
			else
				output += " " + String(value);
			
			output += "\n";
			
			this.output += output;
		}
		
		private function swfLoadComplete(event:Event):void
		{
			var time:uint = getTimer();
			var data:ByteArray = this.loader.data;
			this.console.text = "File loaded (" + this.request.url + "): " + this.loader.bytesTotal + " bytes\n";
			
			data.endian = Endian.LITTLE_ENDIAN;
			var format:String = data.readUTFBytes(3);
			var compressed:Boolean = format == "CWS";
			if(format == "FWS" || format == "CWS")
			{
				this.log("Signature", format, 0);
				this.log("Version", data.readUnsignedByte(), 0);
				this.log("FileLength", data.readUnsignedInt() + " bytes, uncompressed", 0);
			}
			else
			{
				this.console.appendText("INVALID FILE HEADER");
				return;
			}
			data.readBytes(data);
			data.length -= 8;
			if(compressed)
				data.uncompress();
			
			data.position = 0;
			var stageRectangle:Rectangle = this.readRectangle(data);
			this.log("Width", stageRectangle.width + " px", 0);
			this.log("Height", stageRectangle.height + " px", 0);
			this.log("FrameSize", data.readUnsignedByte(), 0);
			this.log("FrameRate", data.readUnsignedByte() + " fps", 0);
			this.log("FrameCount", data.readUnsignedShort(), 0);
			
			while(data.bytesAvailable)
				this.readSWFTag(data);
			
			this.stage.addEventListener(MouseEvent.CLICK, this.clickHandler, false, 0, true);
			
			this.console.appendText("Log generated (" + this.output.length + " bytes in " + (getTimer() - time) + " ms)\n\nClick to copy in clipboard\n");
		}
		
		private function clickHandler(event:MouseEvent):void
		{
			System.setClipboard(this.output);
		}

		// read SWF tag and call handler if present
		private function readSWFTag(data:ByteArray):void
		{
			var tag:uint = data.readUnsignedShort();
			var id:int = tag >>> 6;
			var size:int = tag & 0x3F;
			if(size == 0x3F)
				size = data.readUnsignedInt();
			
			this.log(SWF_TAGS_NAME[id], null, 0);
			/*
			if (this.handlers[id] != null)
			{
				var tagDump:ByteArray = new ByteArray();
				if (size != 0)
					data.readBytes(tagDump, 0, size);
				this.handlers[id](tag, id, size, tagDump);
			}
			else
				data.position += size;
			*/
			
			var tagDump:ByteArray = new ByteArray();
			tagDump.endian = Endian.LITTLE_ENDIAN;//bit swapping
			/*
			 * In SWF files, bytes are swapped whenever reading words and dwords such that a 32-bit value B1B2B3B4 is written as B4B3B2B1, and a 16-bit value B1B2 is written as B2B1.
			 * The reason for this is the differences in storage and retrieval between the Mac and PC processors.
			 */
			if (size != 0)
			{
				data.readBytes(tagDump, 0, size);
				switch(id)
				{
					case 69:
					{
						var attributes:uint = tagDump.readUnsignedByte();
						
						this.log("Reserved", (attributes >>> 5 & 0x7), 1);
						this.log("HasMetadata", Boolean(attributes >>> 4 & 0x1), 1);
						this.log("ActionScript3", Boolean(attributes >>> 3 & 0x1), 1);
						this.log("Reserved", (attributes >>> 1 & 0x3), 1);
						this.log("UseNetwork", Boolean(attributes & 0x1), 1);
						this.log("Reserved", (tagDump.readUnsignedByte() << 16 | tagDump.readUnsignedByte() << 8 | tagDump.readUnsignedByte()), 1);
						break;
					}
					case 9:
					{
						this.log("BackgroundColor", "0x" + (tagDump.readUnsignedByte() << 16 | tagDump.readUnsignedByte() << 8 | tagDump.readUnsignedByte()).toString(16), 1);
						break;
					}
					case 82:
					{
						var flags:uint = tagDump.readUnsignedInt();
						this.log("Flags", flags, 1);
						this.log("Name", this.readString(tagDump), 1);
						var ABCDump:ByteArray = new ByteArray();
						tagDump.readBytes(ABCDump, 0, tagDump.bytesAvailable);
						this.log("ABCData", tagDump, 1);
						break;
					}
					case 86:
					{
						var sceneCount:uint = this.readEncodedU32(tagDump);
						this.log("SceneCount", sceneCount, 1);
						for (var i:uint = 0; i < sceneCount; i++)
						{
							this.log("Offset" + (i + 1), this.readEncodedU32(tagDump), 1);
							this.log("Name" + (i + 1), this.readString(tagDump), 1);
						}
						var frameLabelCount:uint = this.readEncodedU32(tagDump);
						this.log("FrameLabelCount", frameLabelCount, 1);
						for (var i:uint = 0; i < frameLabelCount; i++)
						{
							this.log("FrameNum" + (i + 1), this.readEncodedU32(tagDump), 1);
							this.log("FrameLabel" + (i + 1), this.readString(tagDump), 1);
						}
						break;
					}
					case 88:
					{
						this.log("FontID", tagDump.readUnsignedShort(), 1);
						this.log("FontName", this.readString(tagDump), 1);
						this.log("FontCopyright", this.readString(tagDump), 1);
						break;
					}
					case 48:
					case 75:
					{
						this.log("FontID", tagDump.readUnsignedShort(), 1);
						var options:uint = tagDump.readUnsignedByte();
						var fontFlagsHasLayout:Boolean = Boolean(options >>> 7 & 0x1);
						this.log("FontFlagsHasLayout", fontFlagsHasLayout, 1);
						this.log("FontFlagsShiftJIS", Boolean(options >>> 6 & 0x1), 1);
						this.log("FontFlagsSmallText", Boolean(options >>> 5 & 0x1), 1);
						this.log("FontFlagsANSI", Boolean(options >>> 4 & 0x1), 1);
						var fontFlagsWideOffsets:Boolean = Boolean(options >>> 3 & 0x1);
						this.log("FontFlagsWideOffsets", fontFlagsWideOffsets, 1);
						var fontFlagsWideCodes:Boolean = Boolean(options >>> 2 & 0x1)
						this.log("FontFlagsWideCodes", fontFlagsWideCodes, 1);
						this.log("FontFlagsItalic", Boolean(options >>> 1 & 0x1), 1);
						this.log("FontFlagsBold", Boolean(options & 0x1), 1);
						var languageCode:uint = tagDump.readUnsignedByte()
						this.log("LanguageCode", LANGUAGE_CODE[languageCode], 1);
						var fontNameLen:uint = tagDump.readUnsignedByte();
						this.log("FontNameLen", fontNameLen, 1);
						this.log("FontName", tagDump.readUTFBytes(fontNameLen), 1);
						//this.log("FontName", this.readString(tagDump), 1);
						this.numGlyphs = tagDump.readUnsignedShort();
						this.log("NumGlyphs", this.numGlyphs, 1);
						
						var offsetTable:Array = [];//Array of uint
						if (fontFlagsWideOffsets)
							for (var i:uint = 0; i < this.numGlyphs; i++)
								offsetTable.push(tagDump.readUnsignedInt());
						else
							for (var i:uint = 0; i < this.numGlyphs; i++)
								offsetTable.push(tagDump.readUnsignedShort());
						
						//Transform offset (starting to OffsetTable for GlyphShapeTable) into relative offset of GlyphShapeTable (after Tag)
						var firstOffset:uint = 0;
						if (offsetTable.length >= 1)
						{
							firstOffset = offsetTable[0];
							for (var i:uint = 0; i < this.numGlyphs; i++)
								offsetTable[i] -= firstOffset;
						}
						
						this.log("OffsetTable", offsetTable, 1);
						var numBytes:uint;
						if (fontFlagsWideOffsets)
							numBytes = tagDump.readUnsignedInt();
						else
							numBytes = tagDump.readUnsignedShort();
						var codeTableOffset:uint = numBytes - firstOffset;//relatvie offset of codeTable (= total length of GlyphShapeTable)
						this.log("CodeTableOffset", codeTableOffset, 1);
						this.log("GlyphShapeTable", null, 1);
						
						var shapesOffsets:Array = offsetTable.concat(codeTableOffset);//create a liste of offset + next data offset
							
						var shapeDump:BitArray;
						var numFillBits:uint;
						var numLineBits:uint;
						var typeFlag:Boolean;
						var bits:uint;
						var straightFlag:Boolean;
						var numBits:uint;
						var generalLineFlag:Boolean;
						var vertLineFlag:Boolean;
						var deltaX:Number;
						var deltaY:Number;
						var controlDeltaX:Number;
						var controlDeltaY:Number;
						var anchorDeltaX:Number;
						var anchorDeltaY:Number;
						var stateNewStyles:Boolean;
						var stateLineStyle:Boolean;
						var stateFillStyle1:Boolean;
						var stateFillStyle0:Boolean;
						var stateMoveTo:Boolean;
						var moveBits:uint;
						var moveDeltaX:Number;
						var moveDeltaY:Number;
						var fillStyleCount:uint;
						
						for (var i:uint = 0; i < this.numGlyphs; i++)
						{
							//extracting glyph data (shape) here
							shapeDump = new BitArray();
							shapeDump.endian = Endian.LITTLE_ENDIAN;//bit swapping
							tagDump.readBytes(shapeDump, 0, uint(shapesOffsets[i + 1] - shapesOffsets[i]));
							
							this.log("ShapeRecord", null, 2);
							numFillBits = shapeDump.readUnsignedBits(4);
							this.log("NumFillBits", numFillBits, 3);
							numLineBits = shapeDump.readUnsignedBits(4);
							this.log("NumLineBits", numLineBits, 3);
							/*
							 * EM square = 1024 x 1024
							 *
							 */
							
							while (true)
							{
								typeFlag = shapeDump.readBoolean();
								if (!typeFlag)//if 0 is EndShapeRecord || StyleChangeRecord
								{
									bits = shapeDump.readUnsignedBits(5);
									if (bits == 0)//next bits are 0 (5 bits + n, where n is number of bits for filling to the end of shape record)
									{
										this.log("EndShapeRecord", null, 3);
										this.log("TypeFlag", typeFlag, 4);
										break;
									}
									else//StyleChangeRecord
									{
										shapeDump.position -= 5;
										this.log("StyleChangeRecord", null, 3);
										this.log("TypeFlag", typeFlag, 4);
										stateNewStyles = shapeDump.readBoolean();
										this.log("StateNewStyles", stateNewStyles, 4);
										stateLineStyle = shapeDump.readBoolean();
										this.log("StateLineStyle", stateLineStyle, 4);
										stateFillStyle1 = shapeDump.readBoolean();
										this.log("StateFillStyle1", stateFillStyle1, 4);
										stateFillStyle0 = shapeDump.readBoolean();
										this.log("StateFillStyle0", stateFillStyle0, 4);
										stateMoveTo = shapeDump.readBoolean();
										this.log("StateMoveTo", stateMoveTo, 4);
										if (stateMoveTo)
										{
											moveBits = shapeDump.readUnsignedBits(5);
											this.log("MoveBits", moveBits, 4);
											moveDeltaX = shapeDump.readBits(moveBits) / 20;
											this.log("MoveDeltaX", moveDeltaX, 4);
											moveDeltaY = shapeDump.readBits(moveBits) / 20;
											this.log("MoveDeltaY", moveDeltaY, 4);
										}
										if (stateFillStyle0)
											this.log("FillStyle0", shapeDump.readUnsignedBits(numFillBits), 4);
										if (stateFillStyle1)
											this.log("FillStyle1", shapeDump.readUnsignedBits(numFillBits), 4);
										if (stateLineStyle)
											this.log("LineStyle", shapeDump.readUnsignedBits(numLineBits), 4);
										if (stateNewStyles)
										{
											this.log("FillStyles", "WARNING NOT SUPPORTED YET!", 4);
										}
									}
								}
								else
								{
									straightFlag = shapeDump.readBoolean();
									numBits = shapeDump.readUnsignedBits(4) + 2;//real number is always 2 more than read value
									if (straightFlag)//StraightEdgeRecord
									{
										this.log("StraightEdgeRecord", null, 3);
										this.log("TypeFlag", typeFlag, 4);
										this.log("StraightFlag", straightFlag, 4);
										this.log("NumBits", numBits, 4);
										generalLineFlag = shapeDump.readBoolean();
										this.log("GeneralLineFlag", generalLineFlag, 4);
										vertLineFlag = false;//TMP
										if (!generalLineFlag)
										{
											vertLineFlag = shapeDump.readBoolean();
											this.log("VertLineFlag", vertLineFlag, 4);
										}
										if (generalLineFlag || !vertLineFlag)
										{
											deltaX = shapeDump.readBits(numBits) / 20;
											this.log("DeltaX", deltaX, 4);
										}
										if (generalLineFlag || vertLineFlag)
										{
											deltaY = shapeDump.readBits(numBits) / 20;
											this.log("DeltaY", deltaY, 4);
										}
									}
									else//CurveEdgeRecord
									{
										this.log("CurveEdgeRecord", null, 3);
										this.log("TypeFlag", typeFlag, 4);
										this.log("StraightFlag", straightFlag, 4);
										this.log("NumBits", numBits, 4);
										controlDeltaX = shapeDump.readBits(numBits) / 20;
										this.log("ControlDeltaX", controlDeltaX, 4);
										controlDeltaY = shapeDump.readBits(numBits) / 20;
										this.log("ControlDeltaY", controlDeltaY, 4);
										anchorDeltaX = shapeDump.readBits(numBits) / 20;
										this.log("AnchorDeltaX", anchorDeltaX, 4);
										anchorDeltaY = shapeDump.readBits(numBits) / 20;
										this.log("AnchorDeltaY", anchorDeltaY, 4);
									}
								}
							}
							//END WHILE
						}
						
						//tagDump.position += codeTableOffset;
						var codeTable:Array = [];//Array of uint
						for (var i:uint = 0; i < this.numGlyphs; i++)
							codeTable.push(tagDump.readUnsignedShort());
						var charTable:Array = [];//Array of char (String)
						for (var i:uint = 0; i < this.numGlyphs; i++)
							charTable.push(String.fromCharCode(codeTable[i]));
						this.log("CodeTable", charTable, 1);
						
						if (fontFlagsHasLayout)
						{
							this.log("FontAscent", tagDump.readShort(), 1);
							this.log("FontDescent", tagDump.readShort(), 1);
							this.log("FontLeading", tagDump.readShort(), 1);
							var fontAdvanceTable:Array = [];//Array of int
							for (var i:uint = 0; i < this.numGlyphs; i++)
								fontAdvanceTable.push(tagDump.readShort());
							this.log("FontAdvanceTable", fontAdvanceTable, 1);
							var fontBoundsTable:Array = [];//Array of Rectangle
							for (var i:uint = 0; i < this.numGlyphs; i++)
								fontBoundsTable.push(this.readRectangle(tagDump));
							this.log("FontBoundsTable", fontBoundsTable, 1);
							var kerningCount:uint = tagDump.readUnsignedShort();
							this.log("KerningCount", kerningCount, 1);
							if (kerningCount >= 1)
							{
								this.log("FontKerningTable", null, 1);
								if(fontFlagsWideCodes)
									for (var i:uint = 0; i < kerningCount; i++)
									{
										this.log("KerningRecord", null, 2);
										this.log("FontKerningCode1", tagDump.readUnsignedShort(), 3);
										this.log("FontKerningCode2", tagDump.readUnsignedShort(), 3);
										this.log("FontKerningAdjustment", tagDump.readShort(), 3);
									}
								else
									for (var i:uint = 0; i < kerningCount; i++)
									{
										this.log("KerningRecord", null, 2);
										this.log("FontKerningCode1", tagDump.readUnsignedByte(), 3);
										this.log("FontKerningCode2", tagDump.readUnsignedByte(), 3);
										this.log("FontKerningAdjustment", tagDump.readShort(), 3);
									}
							}
						}
						
						break;
					}
					case 73:// Is alway after DefineFont 2, 3 (1 ?)
					{
						this.log("FontID", tagDump.readUnsignedShort(), 1);
						var bits:uint = tagDump.readUnsignedByte();
						this.log("CSMTableHint", (bits >>> 6), 1);
						this.log("Reserved", (bits & 0x3F), 1);
						this.log("ZoneTable", null, 1);
						var numZoneData:uint;
						for (var i:uint = 0; i < this.numGlyphs; i++)
						{
							this.log("ZoneRecord", null, 2);//Always 2
							numZoneData = tagDump.readUnsignedByte();
							this.log("NumZoneData", numZoneData, 3);//Always 2
							this.log("ZoneData", null, 3);
							for (var y:uint = 0; y < numZoneData; y++)
							{
								this.log("AlignmentCoordinate", this.readFloat16(tagDump), 4);
								this.log("Range", this.readFloat16(tagDump), 4);
							}
							bits = tagDump.readUnsignedByte();
							this.log("Reserved", (bits >>> 2), 3);
							this.log("ZoneMaskY", (bits >>> 1 & 0x1), 3);
							this.log("ZoneMaskX", (bits & 0x1), 3);
						}
						break;
					}
					case 76:
					{
						var num:uint = tagDump.readUnsignedShort();
						this.log("NumSymbols", num, 1);
						
						for (var i:uint = 0; i < num; i++)
						{
							this.log("Tag" + (i + 1), tagDump.readUnsignedShort(), 2);
							this.log("Name" + (i + 1), this.readString(tagDump), 2);
						}
						break;
					}
				}
			}
		}
		
		/*
		 * Read UTF8 String ended by 8bit null char (0)
		 */
		private function readString(data:ByteArray):String
		{
			var string:String = "";
			var char:String;
			while (data.bytesAvailable)
			{
				char = data.readUTFBytes(1);
				data.position --;
				if (data.readUnsignedByte() == 0)
					break;
				string += char;
			}
			return string;
		}

		/*
		 * read rectangle record (RECT: Nbits, Xmin, Xmax, Ymin, Ymax)
		 * return a Rectangle object representation of data
		 */
		private function readRectangle(data:ByteArray):Rectangle {
			var rectangle:Rectangle = new Rectangle();
			var current:uint = data.readUnsignedByte();
			var size:uint = current >>> 3;// Unsigned Bit stored on 5 bits
			var off:int = 3;
			var read:Function = function():Number
			{
				var value:uint = current << (32 - off) >>> (32 - size);
				off -= size;
				while (off < 0)
				{
					current = data.readUnsignedByte();
					value |= off < -8 ? current << (-off -8) : current >>> (-off -8);
					off += 8;
				}
				return value / 20;//values are mutliplied by 20 (1 twip = 1/20 logical pixel)
			};
			
			rectangle.x = read();
			rectangle.right = read();
			rectangle.y = read();
			rectangle.bottom = read();
			
			return rectangle;
		}
		
		private function readEncodedU32(data:ByteArray):uint
		{
			var result:uint = data.readUnsignedByte();
			if (!(result & 0x00000080))
			{
				//pos++;
				return result;
			}
			result = (result & 0x0000007f) | data.readUnsignedByte() << 7;
			if (!(result & 0x00004000))
			{
				//pos += 2;
				return result;
			}
			result = (result & 0x00003fff) | data.readUnsignedByte() << 14;
			if (!(result & 0x00200000))
			{
				//pos += 3;
				return result;
			}
			result = (result & 0x001fffff) | data.readUnsignedByte() << 21;
			if (!(result & 0x10000000))
			{
				//pos += 4;
				return result;
			}
			result = (result & 0x0fffffff) | data.readUnsignedByte() << 28;
			//pos += 5;
			
			return result;
		}
		
		public function readFloat16(data:ByteArray):Number
		{
			var bits:uint = data.readUnsignedShort();
			var sign:int = bits & 0x8000 ? -1 : 1;
			var exp:int = (bits >> 10) & 0x1f;
			var fraction:Number = bits & 0x3ff;
			if (exp == 0)
			{
				if (fraction == 0)
					return 0;
				else
					return fraction;
			}
			if (exp == 0x1f)
			{
				if (fraction == 0)
					return Number.POSITIVE_INFINITY;
				else
					return Number.NaN;
			}
			return sign * Math.pow(2, exp - 16) * (1 + fraction / 0x400);
		}

	}
}